use std::collections::{BinaryHeap, HashMap, VecDeque};
use std::fmt::Display;
use std::fs::File;
use std::io::{BufRead, BufReader, Lines};
use std::time::Instant;

use lazy_regex::lazy_regex;

// BOILERPLATE
type InputIter = Lines<BufReader<File>>;

fn get_input() -> InputIter {
    let f = File::open("input").unwrap();
    let br = BufReader::new(f);
    br.lines()
}

fn main() {
    let start = Instant::now();
    let ans1 = problem1(get_input());
    let duration = start.elapsed();
    println!("Problem 1 solution: {} [{}s]", ans1, duration.as_secs_f64());

    let start = Instant::now();
    let ans2 = problem2(get_input());
    let duration = start.elapsed();
    println!("Problem 2 solution: {} [{}s]", ans2, duration.as_secs_f64());
}

// PARSE

const MODULE_PATTERN: lazy_regex::Lazy<lazy_regex::Regex> = lazy_regex!("^([%&]?)([a-z]+) -> ([a-z, ]+)$");

#[derive(Debug, PartialEq, Eq, Clone, Copy)]
enum PulseType {
    Low,
    High,
}

impl From<bool> for PulseType {
    fn from(value: bool) -> Self {
        match value {
            true => PulseType::High,
            false => PulseType::Low,
        }
    }
}

impl From<PulseType> for bool {
    fn from(value: PulseType) -> Self {
        match value {
            PulseType::High => true,
            PulseType::Low => false,
        }
    }
}

impl Display for PulseType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            PulseType::Low => f.write_str("low"),
            PulseType::High => f.write_str("high"),
        }
    }
}

#[derive(Debug, Clone, Copy)]
enum ModuleKind {
    FlipFlop,
    Conjunction,
    Broadcast,
    Button,
}

#[derive(Debug, Clone)]
struct Module {
    name: String,
    kind: Option<ModuleKind>,
    inputs: Vec<String>,
    outputs: Vec<String>,
    on: bool,
    last_input_states: HashMap<String, PulseType>,
}

impl Module {
    fn new(name: String) -> Self {
        Self {
            name,
            kind: None,
            inputs: Vec::new(),
            outputs: Vec::new(),
            on: false,
            last_input_states: HashMap::new(),
        }
    }
    fn with_kind(name: String, kind: ModuleKind) -> Self {
        let mut m = Self::new(name);
        m.kind = Some(kind);
        m
    }
    fn and_destinations(mut self, destinations: &str) -> Self {
        println!("{} destinations: {}", self.name, destinations);
        for dest in destinations.split(", ") {
            println!("  {}", dest);
            self.outputs.push(dest.into());
        }
        self
    }
    fn and_source(mut self, source: String) -> Self {
        self.add_source(source);
        self
    }
    fn add_source(&mut self, source: String) {
        self.inputs.push(source.clone());
        self.last_input_states.insert(source, PulseType::Low);
    }
}

impl From<&str> for Module {
    fn from(s: &str) -> Self {
        println!("s: {}", s);
        let Some(re_result) = MODULE_PATTERN.captures(s) else {
            panic!("unparseable module: {}", s);
        };
        let (_, [kind, name, destinations]) = re_result.extract();
        if name == "broadcaster" {
            return Module::with_kind(name.into(), ModuleKind::Broadcast).and_destinations(destinations);
        };
        match kind {
            "%" => Module::with_kind(name.into(), ModuleKind::FlipFlop).and_destinations(destinations),
            "&" => Module::with_kind(name.into(), ModuleKind::Conjunction).and_destinations(destinations),
            _ => panic!("invalid module kind {}", kind),
        }
    }
}

#[derive(Debug)]
struct Job {
    signal: PulseType,
    targets: Vec<String>,
    from: String,
    priority: usize,
}

impl PartialEq for Job {
    fn eq(&self, other: &Self) -> bool {
        self.priority == other.priority
    }
}
impl Eq for Job {}

impl PartialOrd for Job {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        other.priority.partial_cmp(&self.priority)
    }
}

impl Ord for Job {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.partial_cmp(other).unwrap()
    }
}

#[derive(Debug)]
struct Machine {
    modules: HashMap<String, Module>,
    work_queue: BinaryHeap<Job>,
    count_low: u64,
    count_high: u64
}

impl<T: BufRead> From<Lines<T>> for Machine {
    fn from(mut lines: Lines<T>) -> Self {
        let mut modules = HashMap::from([(
            "button".into(),
            Module::with_kind("button".into(), ModuleKind::Button).and_destinations("broadcaster"),
        )]);
        while let Some(Ok(line)) = lines.next() {
            let mut module = Module::from(line.as_str());
            for output in &module.outputs {
                if let Some(output_mod) = modules.get_mut(output) {
                    // already exists, push to input list
                    output_mod.add_source(module.name.to_owned());
                } else {
                    // forward declaration of 'unknown' modules
                    modules.insert(
                        output.clone(),
                        Module::new(output.clone()).and_source(module.name.clone()),
                    );
                }
            }
            if let Some(existing_mod) = modules.get(&module.name) {
                module.inputs = existing_mod.inputs.clone(); // might have been pre-prepared, the rest we take ours
                module.last_input_states = existing_mod.last_input_states.clone();
            }
            modules.insert(module.name.clone(), module);
        }

        Machine {
            modules,
            work_queue: BinaryHeap::new(),
            count_low: 0, count_high: 0
        }
    }
}

impl Machine {
    fn press_button(&mut self) {
        self.send_pulse(PulseType::Low, "broadcaster", "button", 1);
    }
    fn send_pulse(&mut self, signal: PulseType, target: &str, from: &str, priority: usize) {
        // count pulse when it is received
        match signal {
            PulseType::Low => self.count_low += 1,
            PulseType::High => self.count_high += 1,
        };
        match self.modules[target].kind {
            Some(ModuleKind::Button) | Some(ModuleKind::Broadcast) => {
                self.send_all_outputs(PulseType::Low, target, priority)
            }
            Some(ModuleKind::FlipFlop) => match signal {
                PulseType::High => (),
                PulseType::Low => {
                    let new_state = !self.modules.get_mut(target).unwrap().on;
                    self.modules.get_mut(target).unwrap().on = new_state;
                    self.send_all_outputs(new_state.into(), target, priority)
                }
            },
            Some(ModuleKind::Conjunction) => {
                self.modules
                    .get_mut(target)
                    .unwrap()
                    .last_input_states
                    .insert(from.to_owned(), signal);
                if self.modules[target]
                    .last_input_states
                    .values()
                    .all(|state| *state == PulseType::High)
                {
                    self.send_all_outputs(PulseType::Low, target, priority)
                } else {
                    self.send_all_outputs(PulseType::High, target, priority)
                }
            }
            None => (),
        }
    }
    fn send_all_outputs(&mut self, signal: PulseType, from: &str, priority: usize) {
        self.work_queue.push(Job {
            signal,
            targets: self.modules[from].outputs.iter().map(|op| op.to_owned()).collect(),
            from: from.to_owned(),
            priority: priority + 1,
        })
    }
    fn run(&mut self) {
        while let Some(job) = self.work_queue.pop() {
            for target in job.targets {
                self.send_pulse(job.signal, &target, &job.from, job.priority);
            }
        }
    }
}

// PROBLEM 1 solution
const PROBLEM1_ITERATIONS: usize = 1000;
fn problem1<T: BufRead>(input: Lines<T>) -> u64 {
    let mut machine = Machine::from(input);
    println!("{:?}", machine);
    for _ in 0..PROBLEM1_ITERATIONS {
        machine.press_button();
        machine.run();
    }
    machine.count_low * machine.count_high
}

// PROBLEM 2 solution
fn problem2<T: BufRead>(input: Lines<T>) -> u64 {
    0
}

#[cfg(test)]
mod tests {
    use crate::*;
    use std::io::Cursor;

    const EXAMPLE: &str = &"broadcaster -> a
%a -> inv, con
&inv -> b
%b -> con
&con -> output";

    #[test]
    fn problem1_example() {
        let c = Cursor::new(EXAMPLE);
        assert_eq!(problem1(c.lines()), 11687500);
    }

    #[test]
    fn problem2_example() {
        let c = Cursor::new(EXAMPLE);
        assert_eq!(problem2(c.lines()), 0);
    }
}
